package codebreaking

enum GuessResult[+Word]:
  case StillGuessing(nextGuess: Word)
  case FoundSolution(solution: Word)
  case NoSolution extends GuessResult[Nothing]
import GuessResult.*

case class ScoredGuess[Word, Score](guess: Word, score: Score)

abstract class Puzzle[Word, Score]:
  def score(guess: Word, secret: Word): Score

  lazy val allSecrets: Iterable[Word]

  def remainingSecrets(previousGuesses: List[ScoredGuess[Word, Score]]): Iterable[Word] =
    allSecrets.filter: s =>
      previousGuesses.forall(g => score(g.guess, s) == g.score)

  val pickGuess = guessRandom
  def guess(previousGuesses: List[ScoredGuess[Word, Score]]): GuessResult[Word] =
    val remaining = remainingSecrets(previousGuesses)
    if remaining.isEmpty then NoSolution
    else if remaining.tail.isEmpty then FoundSolution(remaining.head)
    else StillGuessing(pickGuess(remaining))

  def guessFirst(remainingWords: Iterable[Word]) =
    remainingWords.head

  val rng = scala.util.Random(0)
  def guessRandom(remainingWords: Iterable[Word]) =
    rng.shuffle(remainingWords).head

  lazy val validGuesses: Iterable[Word]

  case class GroupStats(groupCount: Int, maxGroupSize: Int):
    def score =
      if maxGroupSize == 1 then
        (Int.MaxValue, 0) // Guaranteed to find next round
      else (groupCount, -maxGroupSize)

  def guessByGroups(possibleGuesses: Iterable[Word], remainingWords: Iterable[Word]): Word =
    require(!remainingWords.isEmpty)
    val wordList = remainingWords.toSet
    def groupStats(guess: Word) =
      val groups = wordList.groupBy(score(guess, _))
      GroupStats(groups.size, groups.view.map(_._2.size).max)
    val byGroupStats = possibleGuesses.groupBy(groupStats)
    val bestGuess = byGroupStats.maxBy(_._1.score)
    val bestWord = byGroupStats.view
      .mapValues(_.filter(wordList.contains)).filter(!_._2.isEmpty)
      .maxBy(_._1.score)
    if bestWord._1 == bestGuess._1 then bestWord._2.head
    else bestGuess._2.head

  def guessBest(remainingWords: Iterable[Word]) =
    guessByGroups(validGuesses, remainingWords)

def iterWords[T](letters: Iterable[T], wordLength: Int): LazyList[List[T]] =
  if wordLength == 0 then LazyList(List())
  else
    for
      w <- iterWords(letters, wordLength - 1)
      l <- letters.to(LazyList)
    yield l :: w
